################################################################
# __sync
################################################################
SYNC_TABLE_NAME = "__sync"

SYNC_TABLE = {
    "table_name": SYNC_TABLE_NAME,
    "columns": [
        {
            "column_name": "Id",
            "column_type": "text",
            "column_data": {},
            "column_width": 200,
        },
        {
            "column_name": "Table",
            "column_type": "text",
            "column_data": {},
            "column_width": 120,
        },
        {
            "column_name": "View",
            "column_type": "text",
            "column_data": {},
            "column_width": 120,
        },
        {
            "column_name": "Managers",
            "column_type": "collaborator",
            "column_data": {},
            "column_width": 240,
        },
        {
            "column_name": "LastUpdated",
            "column_type": "date",
            "column_data": {"format": "YYYY-MM-DD HH:mm"},
            "column_width": 120,
        },
        {
            "column_name": "LastSync",
            "column_type": "date",
            "column_data": {"format": "YYYY-MM-DD HH:mm"},
            "column_width": 120,
        },
        {
            "column_name": "Sync",
            "column_type": "button",
            "column_data": {
                "button_name": "Sync",
                "button_color": "#FFFCB5",
                "button_action_list": [
                    {
                        "action_type": "send_notification",
                        "msg": "hello",
                        "to_users": ["woojin.cho@gmail.com"],
                        "user_column": "Managers",
                    }
                ],
            },
            "column_width": 120,
        },
        {
            "column_name": "Log",
            "column_type": "long-text",
            "column_data": {},
            "column_width": 360,
        },
    ],
}

SYNC_TABLE_RESIZE = list()
for c in SYNC_TABLE["columns"]:
    new_column_width = c.pop("column_width")
    SYNC_TABLE_RESIZE.append(
        {"table_name": SYNC_TABLE_NAME, "column_name": c["column_name"], "new_column_width": new_column_width}
    )



import io
from typing import Annotated

from fastapi import Depends, FastAPI

from . import router
from .conf import AWS_S3_BUCKET_NAME, DEV, PROD
from .util import generate_obj_key

app = FastAPI(title="FASTO API")
app.include_router(router.user.router)
app.include_router(router.api_token.router)


################################################################
# Endpoints
################################################################
@app.get("/hello", tags=["System"])
async def hello():
    return {"hello": "world!"}


@app.get("/info/s3", tags=["Info"])
async def info_detination_s3():
    format = "<format>"
    workspace_name = "<workspace>"
    base_name = "<base>"
    table_name = "<table>"
    view_name = "<view>"
    group = "<group>"
    examples = [
        (True, view_name, None),
        (False, view_name, None),
        (True, view_name, group),
        (False, view_name, group),
    ]
    return {
        "bucket": AWS_S3_BUCKET_NAME,
        **{
            f"key {'w/' if group else 'w/o'} view group for {PROD if prod else DEV}": generate_obj_key(
                format=format,
                prod=True,
                workspace_name=workspace_name,
                base_name=base_name,
                table_name=table_name,
                view_name=view_name,
                group=group,
            )
            for prod, view_name, group, in examples
        },
    }



import asyncio
import logging
from datetime import datetime
from typing import List, Union

import aiohttp
import orjson
from pydantic import BaseModel
from tabulate import tabulate

from ..model import Admin, ApiToken, Base, BaseToken, Column, File, Table, Team, User, Webhook
from .conf import SEATABLE_URL

logger = logging.getLogger()

TABULATE_CONF = {"tablefmt": "psql", "headers": "keys"}


def parse_name(*name, delim: str = "/"):
    return [x for e in name for x in (e.split(delim) if e else [None])][: len(name)]


################################################################
# HttpClient
################################################################
class HttpClient:
    def __init__(self, seatable_url: str = SEATABLE_URL):
        self.seatable_url = seatable_url.rstrip("/")

        self.headers = {"accept": "application/json"}
        self.debug = False
        self._request = None

    async def info(self):
        async with self.session_maker() as session:
            return await self.request(session=session, method="GET", url="/server-info/")

    async def ping(self):
        async with self.session_maker() as session:
            return await self.request(session=session, method="GET", url="/api2/ping/")

    def session_maker(self, token: str = None):
        headers = self.headers.copy()
        if token:
            headers.update({"authorization": "Bearer {}".format(token)})
        return aiohttp.ClientSession(base_url=self.seatable_url, headers=headers)

    async def request(
        self,
        session: aiohttp.ClientSession,
        method: str,
        url: str,
        json: str = None,
        data: bytes = None,
        **params,
    ):
        # for debug
        self._request = {
            "method": method,
            "url": url,
            "json": json if not json else {k: v for k, v in json.items() if v},
            "data": data,
            "params": params if not params else {k: v for k, v in params.items() if v},
        }

        async with session.request(**self._request) as response:
            response.raise_for_status()

            if self.debug:
                print(response.headers)
                return await response.content()
            try:
                if response.content_type in ["application/json"]:
                    return await response.json()
                if response.content_type in ["text/html"]:
                    logger.warning(f"! content-type: {response.content_type}")
                    body = await response.text()
                    return orjson.loads(body)
                if response.content_type in [
                    "application/ms-excel",
                    "application/x-zip-compressed",
                ]:
                    content = b""
                    async for data in response.content.iter_chunked(2048):
                        content += data
                    if len(content) != response.content_length:
                        raise ValueError()
                    return File(filename=response.content_disposition.filename, content=content)

            except Exception as ex:
                raise ex

    @staticmethod
    def print(records: List[dict], tabulate_conf: dict = TABULATE_CONF):
        print(tabulate(records, **tabulate_conf))



from datetime import datetime
from enum import Enum
from typing import Any, List, Union

import orjson
from pydantic import BaseModel, Extra, Field, validator

__all__ = [
    "DTABLE_ICON_LIST",
    "DTABLE_ICON_COLORS",
    "ColumnType",
    "Column",
    "File",
    "Table",
    "Metadata",
    "View",
    "SharedView",
    "User",
    "UserInfo",
    "Base",
    "BaseActivity",
    "BaseInfo",
    "Group",
    "Workspace",
    "BaseExternalLink",
    "ViewExternalLink",
]


DTABLE_ICON_LIST = [
    "icon-worksheet",
    "icon-task-management",
    "icon-software-test-management",
    "icon-design-assignment",
    "icon-video-production",
    "icon-market-analysis",
    "icon-data-analysis",
    "icon-product-knowledge-base",
    "icon-asset-management",
    "icon-financial-information-record",
    "icon-dollar",
    "icon-company-inventory",
    "icon-customer-inquiry",
    "icon-customer-list",
    "icon-product-list",
    "icon-store-address",
    "icon-leave-record",
    "icon-administrative-matters-calendar",
    "icon-customer-relationship",
    "icon-teachers-list",
    "icon-book-library",
    "icon-server-management",
    "icon-time-management",
    "icon-work-log",
    "icon-online-promotion",
    "icon-research",
    "icon-user-interview",
    "icon-client-review",
    "icon-club-members",
]

DTABLE_ICON_COLORS = [
    "#FF8000",
    "#FFB600",
    "#E91E63",
    "#EB00B1",
    "#7626FD",
    "#972CB0",
    "#1DDD1D",
    "#4CAF50",
    "#02C0FF",
    "#00C9C7",
    "#1688FC",
    "#656463",
]


class _Model(BaseModel):  # remove "extra=Extra.forbid" here
    @validator("*", pre=True)
    def empty_to_none(cls, v):
        if v == "":
            return None
        return v


class ColumnType(Enum):
    text = "text"
    long_text = "long-text"
    number = "number"
    collaborator = "collaborator"
    date = "date"
    duration = "duration"
    single_select = "single-select"
    multiple_select = "multiple-select"
    image = "image"
    file = "file"
    email = "email"
    url = "url"
    checkbox = "checkbox"
    rating = "rating"
    formula = "formula"
    link = "link"
    link_formula = "link-formula"
    creator = "creator"
    ctime = "ctime"
    last_modifier = "last-modifier"
    mtime = "mtime"
    auto_number = "auto-number"


class Column(_Model):
    key: str  # '0000'
    type: str  # 'text' # ColumnType
    name: str = None  # 'Hive DB'
    width: int  # 199
    editable: bool  # True
    resizable: bool  # True
    draggable: bool = None  # True
    frozen: bool = None
    colorbys: dict = None
    formatter: dict = None
    data: Any = None  # None
    permission_type: str = None  # ''
    permitted_users: List[str] = None  # []
    editor: dict = None
    edit_metadata_permission_type: str = None  # ''
    edit_metadata_permitted_users: List[str] = None  # []
    description: str = None  # None
    # [NOTE] 아래 prop들은 컬럼 삭제하였다가 undo하면 이후부터 생김...
    idx: int = None
    last_frozen: bool = None
    left: int = None
    rowType: str = None

    def to_column_info(self):
        return {
            "column_name": self.name,
            "column_type": self.type,
            "column_data": self.data,
        }


class View(_Model):
    id: str = Field(..., alias="_id")  # '0000'
    name: str  # 'Default View'
    type: str  # 'table'
    private_for: str = None  # None
    is_locked: bool = None  # False
    row_height: str = None  # 'default'
    filter_conjunction: str = None  # 'And'
    filters: List[dict] = None  # []
    sorts: List[dict] = None  # []
    groupbys: List[dict] = None  # []
    colorbys: dict = None  # {}
    hidden_columns: List[str] = None  # []
    rows: List[str] = None  # []
    formula_rows: dict = None  # {}
    link_rows: dict = None  # {}
    summaries: dict = None  # {}
    colors: dict = None  # {}
    column_colors: dict = None  # {}
    groups: List[str] = None  # []


class SharedView(_Model):
    id: int
    shared_name: str
    workspace_id: int
    dtable_name: str
    table_id: str  # '0000'
    view_id: str  # '0000'
    from_user: str
    to_user: str
    from_user_name: str
    to_user_name: str
    permission: str
    color: str = None
    text_color: str = None
    icon: str = None


class Table(_Model):
    id: str = Field(..., alias="_id")
    name: str
    is_header_locked: bool = None
    columns: List[Column]
    views: List[View] = None
    summary_configs: dict = None
    header_settings: dict = None

    def to_table_info(self):
        return {
            "table_name": self.name,
            "columns": [c.to_column_info() for c in self.columns],
        }


class Metadata(_Model):
    tables: List[Table]
    version: int
    format_version: int
    scripts: List[dict] = None
    settings: dict = None


class User(_Model):
    email: str  # '2926d3fa3a364558bac8a550811dbe0e@auth.local'
    name: str  # 'admin'
    name_pinyin: str = None
    contact_email: str  # 'woojin.cho@gmail.com'
    login_id: str = None  # ''
    phone: str = None
    avatar_url: str = None
    unit: str = None  # ''
    id_in_org: str = None  # ''
    is_staff: bool  # True
    is_active: bool  # True
    role: str = None  # 'default'
    permission: str = None  # "r" or "rw"
    workspace_id: int = None  # 1
    create_time: datetime  # '2023-05-21T03:04:26+00:00'
    last_login: datetime = None  # '2023-05-28T11:42:01+00:00'
    storage_quota: int = None
    storage_usage: int  # 0
    row_limit: int = None
    row_usage: int = None
    rows_count: int = None  # 0
    source: str = None
    quota_total: int = None
    quota_usage: int = None


class UserInfo(_Model):
    email: str  # "876543216569491ba42905bf1647fd3f@auth.local"
    name: str
    name_pinyin: str = None
    contact_email: str = None  # "michael@example.com"
    login_id: str = None  # ""
    group_id: str = None  # 1
    id_in_org: str = None  # ''
    is_admin: bool = None  # true
    role: str = None  # "Owner
    permission: str = None  # "r" or "rw"
    avatar_url: str = (
        None  # "https://cloud.seatable.io/image-view/avatars/3/7/a0a57575a3ca0c78e8c5b6b0d0dbda/resized/80/cd7f6edd2c75afd3b7299917b3767c0f.png"
    )


class Base(_Model):
    id: int = None
    workspace_id: int  # 3
    uuid: str  # '166424ad-b023-47a0-9a35-76077f5b629b'
    name: str  # 'employee'
    creator: str = None  # "Jasmin Tee"
    creator_email: str = None
    modifier: str = None  # "Jasmin Tee"
    created_at: datetime  # '2023-05-21T04:33:18+00:00'
    updated_at: datetime  # '2023-05-21T04:33:30+00:00'
    color: str = None  # None
    text_color: str = None  # None
    icon: str = None  # None
    is_encrypted: bool  # False
    in_storage: bool  # True
    org_id: int = None  # -1
    email: str = None  # '1@seafile_group'
    group_id: int = None  # 1
    owner: str = None  # 'Employee (group)'
    owner_deleted: bool = False  # False
    from_user: str = None
    from_user_name: str = None
    from_user_avatar: str = None
    permission: str = None
    file_size: int = None  # 10577
    rows_count: int = None  # 0

    def to_record(self):
        return {
            "base_uuid": self.uuid,
            "workspace_id": self.workspace_id,
            "group_name": self.owner.replace(" (group)", ""),
            "base_name": self.name,
            "created_at": self.created_at.strftime("%Y-%m-%dT%H:%M"),
            "updated_at": self.updated_at.strftime("%Y-%m-%dT%H:%M"),
            "owner_deleted": self.owner_deleted,
            "rows_count": self.rows_count,
        }


class BaseInfo(_Model):
    id: int  # 378
    workspace_id: int  # 504
    uuid: str  # "12345678-3643-489b-880c-51c8ee2a9a20"
    name: str  # "Customers"
    creator: str = None  # "Jasmin Tee"
    modifier: str = None  # "Jasmin Tee"
    created_at: str  # "2020-11-20T11:57:30+00:00"
    updated_at: str  # "2020-11-20T11:57:30+00:00"
    color: str = None  # null
    text_color: str = None  # null
    icon: str = None  # nul
    is_encrypted: bool = None
    in_storage: bool = None
    starred: bool = None
    from_user: str = None
    from_user_name: str = None
    from_user_avatar: str = None
    permission: str = None


class Group(_Model):
    id: int
    name: str
    owner: str
    owner_name: str
    created_at: datetime
    quota: int
    parent_group_id: int
    size: int


class Workspace(_Model):
    id: int = None
    name: str
    type: str
    folders: List[dict] = None
    shared_bases: List[BaseInfo] = Field(None, alias="shared_table_list")
    shared_views: List[View] = Field(None, alias="shared_view_list")
    bases: List[BaseInfo] = Field(None, alias="table_list")
    group_id: int = None
    group_owner: str = None
    group_admins: List[str] = None
    group_member_count: int = None
    group_shared_dtables: List[dict] = None
    group_shared_views: List[dict] = None

    def to_record(self):
        bases = self.bases or self.shared_bases
        return {
            "type": self.type,
            "workspace_id": self.id,
            "workspace": self.name,
            "folders": [x["name"] for x in self.folders] if self.folders else self.folders,
            "bases": [x.name for x in bases] if bases else bases,
        }


class File(_Model):
    filename: str
    content: bytes


class BaseActivity(_Model):
    author: str = None
    app: str = None
    op_id: int
    op_time: datetime
    operation: dict

    @validator("operation", pre=True)
    def load_json(cls, v):
        return orjson.loads(v)


class BaseExternalLink(_Model):
    id: int
    from_dtable: str
    from_base_uuid: str = None
    creator: str
    creator_name: str
    token: str
    permission: str
    create_at: datetime
    view_cnt: int
    url: str


class ViewExternalLink(_Model):
    id: int
    from_dtable: str
    from_base_uuid: str = None
    creator: str
    creator_name: str
    token: str
    permission: str
    create_at: datetime
    view_cnt: int
    table_id: str
    view_id: str
    url: str
    is_custom: bool
    expire_date: datetime = None



from importlib.metadata import PackageNotFoundError, version

from .client import AccountClient, AdminClient, BaseClient, UserClient

try:
    __version__ = version("plantable")
except PackageNotFoundError:
    # package is not installed
    pass



from typing import Annotated

import aioboto3
from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials

from ...client import UserClient
from ..conf import AIOBOTO3_CONF
from ..util import upload_to_s3, view_to_parquet

session = aioboto3.Session(**AIOBOTO3_CONF)
router = APIRouter(prefix="/user", tags=["UserClient"])
security = HTTPBasic()


################################################################
# Basic Auth
################################################################
async def get_user_client(credentials: Annotated[HTTPBasicCredentials, Depends(security)]):
    uc = UserClient(seatable_username=credentials.username, seatable_password=credentials.password)
    try:
        _ = await uc.login()
    except Exception as ex:
        if ex.status == 400:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect username or password",
                headers={"WWW-Authenticate": "Basic"},
            )
        raise ex
    return uc


################################################################
# Endpoints
################################################################
# My SeaTable Account Info
@router.get("/info")
async def info_my_seatable_account(user_client: Annotated[dict, Depends(get_user_client)]):
    return await user_client.get_account_info()


# Export View to S3 Parquet
@router.get("/export/parquet/view")
async def export_view_to_s3_with_parquet(
    user_client: Annotated[dict, Depends(get_user_client)],
    workspace_name: str,
    base_name: str,
    table_name: str,
    view_name: str,
    group: str = None,
    prod: bool = False,
):
    base_client = await user_client.get_base_client_with_account_token(
        workspace_name_or_id=workspace_name, base_name=base_name
    )
    content = await view_to_parquet(client=base_client, table_name=table_name, view_name=view_name)

    return await upload_to_s3(
        session=session,
        content=content,
        format="parquet",
        prod=prod,
        workspace_name=workspace_name,
        base_name=base_name,
        table_name=table_name,
        view_name=view_name,
        group=group,
    )



import aioboto3
from fastapi import APIRouter, Depends
from fastapi.security import APIKeyHeader

from ...client import BaseClient
from ..conf import AIOBOTO3_CONF
from ..util import upload_to_s3, view_to_parquet

session = aioboto3.Session(**AIOBOTO3_CONF)
router = APIRouter(prefix="/api-token", tags=["ApiTokenClient"])

api_key_header = APIKeyHeader(name="Token")


################################################################
# Basic Auth
################################################################
async def get_base_client(api_token: str = Depends(api_key_header)) -> BaseClient:
    bc = BaseClient(api_token=api_token)
    return bc


################################################################
# Endpoints
################################################################
# My API Token
@router.get("/info")
async def info_my_seatable_api_token(
    base_client: BaseClient = Depends(get_base_client),
) -> dict:
    return base_client.base_token


# Export View to S3 Parquet
@router.get("/export/parquet/view")
async def export_view_to_s3_with_parquet(
    workspace_name: str,
    base_name: str,
    table_name: str,
    view_name: str,
    group: str = None,
    prod: bool = False,
    base_client: BaseClient = Depends(get_base_client),
):
    content = await view_to_parquet(client=base_client, table_name=table_name, view_name=view_name)

    return await upload_to_s3(
        session=session,
        content=content,
        format="parquet",
        prod=prod,
        workspace_name=workspace_name,
        base_name=base_name,
        table_name=table_name,
        view_name=view_name,
        group=group,
    )



import asyncio
import logging
from datetime import datetime
from typing import List, Union

import aiohttp
import orjson
from pydantic import BaseModel
from tabulate import tabulate

from ..model import (
    DTABLE_ICON_COLORS,
    DTABLE_ICON_LIST,
    AccountInfo,
    Activity,
    Admin,
    ApiToken,
    Base,
    BaseInfo,
    BaseToken,
    Column,
    File,
    SharedView,
    Table,
    Team,
    User,
    UserInfo,
    Webhook,
    Workspace,
)
from .account import AccountClient
from .core import parse_name

logger = logging.getLogger()


################################################################
# UserClient
################################################################
class UserClient(AccountClient):
    ################################################################
    # Helper
    ################################################################
    async def get_base_client_with_account_token(
        self, workspace_name_or_id: Union[str, int] = None, base_name: str = None
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)
        return await super().get_base_client_with_account_token(workspace_id=workspace.id, base_name=base_name)

    ################################################################
    # USER
    ################################################################
    # [USER] get account info
    async def get_account_info(self, model: BaseModel = AccountInfo):
        METHOD = "GET"
        URL = "/api2/account/info/"

        async with self.session_maker(token=self.account_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL)

        if model:
            results = model(**results)

        return results

    # [USER] update email address
    async def update_email_address(self):
        raise NotImplementedError

    # [USER] upload/update user avatar
    async def update_user_avartar(self):
        raise NotImplementedError

    # [USER] get public user info
    async def get_public_user_info(self, user_id: str, model: BaseModel = UserInfo):
        METHOD = "GET"
        URL = f"/api/v2.1/user-common-info/{user_id}/"

        async with self.session_maker(token=self.account_token) as session:
            results = await self.request(session=session, method=METHOD, url=URL)

        if model:
            results = model(**results)

        return results

    # [USER] list public user info
    async def list_public_user_info(self, user_id_list: List[str], model: BaseModel = UserInfo):
        METHOD = "POST"
        URL = "/api/v2.1/user-list/"
        JSON = {"user_id_list": user_id_list}
        ITEM = "user_list"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    ################################################################
    # BASE
    ################################################################
    # [BASES] list bases user can admin
    async def list_bases(self, model: BaseModel = Base):
        METHOD = "GET"
        URL = "/api/v2.1/user-admin-dtables/"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response

        if model:
            results["personal"] = [model(**x) for x in results["personal"]]
            for group in results["groups"]:
                group["dtables"] = [model(**x) for x in group["dtables"]]

        return results

    # [BASES] Create Base
    async def create_base(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        icon: str = None,
        color: str = None,
        model: BaseModel = Base,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "POST"
        URL = f"/api/v2.1/dtables/"
        ITEM = "table"
        DATA = aiohttp.FormData([("workspace_id", workspace.id), ("name", base_name)])
        _ = DATA.add_field("icon", icon) if icon else None
        _ = DATA.add_field("color", color) if color else None

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, data=DATA)
            results = response[ITEM]

        if model:
            results = model(**results)

        return results

    # [BASES] update base
    async def update_base(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        new_base_name: str = None,
        new_icon: str = None,
        new_color: str = None,
        model: BaseModel = Base,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "PUT"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/"
        ITEM = "table"
        DATA = aiohttp.FormData([("name", base_name)])
        _ = DATA.add_field("new_name", new_base_name) if new_base_name else None
        _ = DATA.add_field("icon", new_icon) if new_icon else None
        _ = DATA.add_field("color", new_color) if new_color else None

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, data=DATA)
            results = response[ITEM]

        if model:
            results = model(**results)

        return results

    # [BASES] delete base
    # NOT WORKING
    async def delete_base(self, workspace_name_or_id: Union[str, int]):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "DELETE"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/"
        ITEM = "success"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        return results

    # [BASES] get base activities
    async def get_base_activities(self, model: BaseModel = Activity):
        METHOD = "GET"
        URL = "/api/v2.1/dtable-activities/"
        ITEM = "table_activities"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # [BASES] list_favorites
    async def list_favorites(self, model: BaseModel = BaseInfo):
        METHOD = "GET"
        URL = "/api/v2.1/starred-dtables/"
        ITEM = "user_starred_dtable_list"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # [BASES] favorite base
    async def favorite_base(self, base: Base):
        METHOD = "POST"
        URL = "/api/v2.1/starred-dtables/"
        ITEM = "success"
        JSON = {"dtable_uuid": base.uuid}

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            results = response[ITEM]

        return results

    # [BASES] unfavorite base
    async def unfavorite_base(self, base: Base):
        METHOD = "DELETE"
        URL = "/api/v2.1/starred-dtables/"
        ITEM = "success"
        PARAMS = {"dtable_uuid": base.uuid}

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, **PARAMS)
            results = response[ITEM]

        return results

    # [BASES] (custom) get base by name
    async def get_base(self, workspace_name: str, base_name: str):
        results = await self.list_bases()

        # personal bases
        if workspace_name == "personal":
            for base in results["personal"]:
                if base.name == base_name:
                    return base

        # group bases
        for group in results["groups"]:
            if group["group_name"] != workspace_name:
                continue
            for base in group["dtables"]:
                if base.name == base_name:
                    return base
        else:
            raise KeyError()

    ################################################################
    # GROUPS & WORKSPACES
    ################################################################
    # [GROUPS & WORKSPACES] list workspaces
    async def list_workspaces(self, detail: bool = True, incl: List[str] = None, model: BaseModel = Workspace) -> dict:
        """
        incl: "personal" or "group"
        """
        METHOD = "GET"
        URL = "/api/v2.1/workspaces/"
        ITEM = "workspace_list"
        PARAMS = {"detail": str(detail).lower()}

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, **PARAMS)
            results = response[ITEM]

        if incl:
            incl = incl if isinstance(incl, list) else [incl]
            results = [x for x in results if x["type"] in incl]

        if model:
            # [NOTE]
            # shared는 데이터 형식이 다르기 때문에 별도 method 사용, 여기서는 ignore
            results = [x for x in results if x["type"] != "shared"]
            results = [model(**x) for x in results]

        return results

    # [GROUPS & WORKSPACES] get workspace
    async def get_workspace(self, name_or_id: Union[str, int], workspace_type: str = "group"):
        """
        workspace_type: "group", "personal", "starred", or "shared"
        """
        workspaces = await self.list_workspaces(detail=True, model=Workspace)
        for workspace in workspaces:
            if workspace_type and workspace.type != workspace_type:
                continue
            if isinstance(name_or_id, str) and workspace.name == name_or_id:
                return workspace
            if isinstance(name_or_id, int) and workspace.id == name_or_id:
                return workspace
        else:
            raise KeyError()

    ################################################################
    # (CUSTOM) LS & GET
    ################################################################
    # (custom) get
    async def get(self, workspace_name_or_id: str, base_name: str = None, table_name: str = None):
        workspace_name_or_id, base_name, table_name = parse_name(workspace_name_or_id, base_name, table_name)

        if table_name:
            bc = await self.get_base_client_with_account_token(
                workspace_name_or_id=workspace_name_or_id, base_name=base_name
            )
            return await bc.get_table(table_name=table_name)
        if base_name:
            return await self.get_base(workspace_name_or_id=workspace_name_or_id, base_name=base_name)
        return await self.get_workspace(name_or_id=workspace_name_or_id)

    # (custom) ls
    async def ls(
        self,
        workspace_name_or_id: str = None,
        base_name: str = None,
        table_name: str = None,
    ):
        # coros
        async def _get_records(workspace_name_or_id, base_name):
            bc = await self.get_base_client_with_account_token(
                workspace_name_or_id=workspace_name_or_id, base_name=base_name
            )
            tables = await bc.list_tables()
            return {
                "base_uuid": base.uuid,
                "base": base_name,
                "tables": [x.name for x in tables],
            }

        # ls workspaces
        if not workspace_name_or_id:
            workspaces = await self.list_workspaces(detail=True, model=Workspace)
            records = [x.to_record() for x in workspaces]
            self.print(records=records)
            return

        # ls bases
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)
        if not base_name:
            list_bases = list()
            if workspace.shared_bases:
                for base in workspace.shared_bases:
                    list_bases.append((workspace.name, base.name))
            if workspace.bases:
                for base in workspace.bases:
                    list_bases.append((workspace.name, base.name))

            records = await asyncio.gather(*[_get_records(w, b) for w, b in list_bases])
            self.print(records=records)
            return

        # ls tables
        bc = await self.get_base_client_with_account_token(
            workspace_name_or_id=workspace_name_or_id, base_name=base_name
        )
        await bc.ls(table_name=table_name)

    # [GROUPS & WORKSPACES] copy base from workspace
    async def copy_base_from_workspace(
        self,
        src_workspace_name_or_id: str,
        src_base_name: str,
        dst_workspace_name_or_id: str,
    ) -> dict:
        src_workspace = await self.get_workspace(workspace_name_or_id=src_workspace_name_or_id)
        dst_workspace = await self.get_workspace(workspace_name_or_id=dst_workspace_name_or_id)

        METHOD = "POST"
        URL = f"/api/v2.1/dtable-copy/"
        JSON = {
            "src_workspace_id": src_workspace.id,
            "name": src_base_name,
            "dst_workspace_id": dst_workspace.id,
        }
        ITEM = "dtable"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            results = response[ITEM]

        return results

    # [GROUPS & WORKSPACES] copy base from external link
    async def copy_base_from_external_link(
        self,
        link: str,
        dst_workspace_name: str,
    ) -> dict:
        dst_workspace = await self.get_workspace(workspace_name=dst_workspace_name)

        METHOD = "POST"
        URL = f"/api/v2.1/dtable-external-link/dtable-copy/"
        JSON = {"link": link, "dst_workspace_id": dst_workspace.id}
        ITEM = "dtable"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            results = response[ITEM]

        return results

    ################################################################
    # ATTACHMENT
    ################################################################
    # TBD

    ################################################################
    # IMPORT & EXPORT
    ################################################################
    # (CUSTOM) Get Table Name and View Name from Table ID and View ID - View ID is Optional
    async def get_ids_by_names(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        table_name: str,
        view_name: str = None,
    ):
        bc = await self.get_base_client_with_account_token(
            workspace_name_or_id=workspace_name_or_id, base_name=base_name
        )
        tables = await bc.list_tables()

        for table in tables:
            if table.name == table_name:
                break
        else:
            KeyError
        if view_name is None:
            return table.id

        for view in table.views:
            if view.name == view_name:
                return table.id, view.id
        else:
            KeyError

    # Import Base from xlsx or csv
    async def import_base_from_xlsx_or_csv(
        self, workspace_name_or_id: Union[str, int], file: bytes, folder_id: int = None
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "POST"
        URL = f"/api/v2.1/workspace/{workspace.id}/synchronous-import/import-excel-csv-to-base/"
        JSON = {"dtable": file, "folder": folder_id}

        raise NotImplementedError

    # Import Table from xlsx or csv
    async def import_table_from_xlsx_or_csv(
        self,
        workspace_name_or_id: Union[str, int],
        file: bytes,
        base_uuid: str,
        table_name: str,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "POST"
        URL = f"/api/v2.1/workspace/{workspace.id}/synchronous-import/import-excel-csv-to-table/"
        JSON = {
            "workspace_id": workspace.id,
            "file": file,
            "dtable_uuid": base_uuid,
            "table_name": table_name,
        }
        raise NotImplementedError

    # Update Table from xlsx or csv
    async def update_base_from_xlsx_or_csv(
        self,
        workspace_name_or_id: Union[str, int],
        file: bytes,
        base_uuid: str,
        table_name: str,
        selected_columns: List[str],
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "POST"
        URL = f"/api/v2.1/workspace/{workspace.id}/synchronous-import/update-table-via-excel-csv/"
        JSON = {
            "workspace_id": workspace.id,
            "file": file,
            "dtable_uuid": base_uuid,
            "table_name": table_name,
            "selected_columns": ",".join(selected_columns) if isinstance(selected_columns, list) else selected_columns,
        }
        raise NotImplementedError

    # Export Base
    # NOT WORKING
    async def export_base(self, workspace_name_or_id: Union[str, int], base_name: str):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace.id}/synchronous-export/export-dtable/"
        PARAMS = {"dtable_name": base_name}

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, **PARAMS)

        return response

    # Export Table
    async def export_table(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        table_id: int = None,
        table_name: str = None,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace.id}/synchronous-export/export-table-to-excel/"
        PARAMS = {
            "dtable_name": base_name,
            "table_id": table_id,
            "table_name": table_name,
        }

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, **PARAMS)

        return response

    # (CUSTOM) Export Table by Name
    async def export_table_by_name(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        table_name: str = None,
    ):
        table_id = await self.get_ids_by_names(
            workspace_name_or_id=workspace_name_or_id,
            base_name=base_name,
            table_name=table_name,
        )
        return await self.export_table(
            workspace_name_or_id=workspace_name_or_id,
            base_name=base_name,
            table_id=table_id,
            table_name=table_name,
        )

    # Export View
    async def export_view(
        self,
        workspace_name_or_id: int,
        base_name: str,
        table_id: str,
        table_name: str,
        view_id: str,
        view_name: str,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace.id}/synchronous-export/export-view-to-excel/"
        PARAMS = {
            "dtable_name": base_name,
            "table_id": table_id,
            "table_name": table_name,
            "view_id": view_id,
            "view_name": view_name,
        }

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, **PARAMS)

        return response

    # (CUSTOM) Export View by Name
    async def export_view_by_name(self, workspace_name_or_id: int, base_name: str, table_name: str, view_name: str):
        table_id, view_id = await self.get_ids_by_names(
            workspace_name_or_id=workspace_name_or_id,
            base_name=base_name,
            table_name=table_name,
            view_name=view_name,
        )

        return await self.export_view(
            workspace_name_or_id=workspace_name_or_id,
            base_name=base_name,
            table_id=table_id,
            table_name=table_name,
            view_id=view_id,
            view_name=view_name,
        )

    ################################################################
    # SHARING
    ################################################################
    # My User View Shares
    async def list_shared_views(self, model: BaseModel = SharedView):
        METHOD = "GET"
        URL = "/api/v2.1/dtables/view-shares-user-shared/"
        ITEM = "view_share_list"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # List User Shares
    async def list_users_share_base(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        model: BaseModel = UserInfo,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/{base_name}/share/"
        ITEM = "user_list"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # Create User Share
    async def create_user_share(self):
        raise NotImplementedError

    # Update User Share
    async def update_user_share(self):
        raise NotImplementedError

    # Delete User Share
    async def delete_user_share(self):
        raise NotImplementedError

    # My Group Shares
    async def list_tables_shared_to_my_groups(self) -> List[dict]:
        METHOD = "GET"
        URL = "/api/v2.1/dtables/group-shared/"
        ITEM = "group_shared_dtables"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        return results

    # List Group Shares
    async def list_tables_shared_to_group(self, workspace_name_or_id: Union[str, int], base_name: str):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)
        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/{base_name}/group-shares/"
        ITEM = "dtable_group_share_list"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        return results

    # Create Group Share
    async def create_group_share(self):
        raise NotImplementedError

    ################################################################
    # SHARING LINKS
    ################################################################
    # TBD

    ################################################################
    # COMMON DATASET
    ################################################################
    # TBD

    ################################################################
    # DEPARTMENTS
    ################################################################
    # TBD

    ################################################################
    # FORMS
    ################################################################
    # TBD

    ################################################################
    # AUTOMATIONS
    ################################################################
    # TBD

    ################################################################
    # NOTIFICATIONS
    ################################################################
    # TBD

    ################################################################
    # SYSTEM NOTIFICATIONS
    ################################################################
    # TBD

    ################################################################
    # E-MAIL ACCOUNTS
    ################################################################
    # TBD

    ################################################################
    # WEBHOOKS
    ################################################################
    # [WEBHOOKS] list webhooks
    async def list_webhooks(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        model: BaseModel = Webhook,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "GET"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/{base_name}/webhooks/"
        ITEM = "webhook_list"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        if model:
            results = [model(**x) for x in results]

        return results

    # [WEBHOOKS] create webhook
    async def create_webhook(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        url: str,
        secret: int = 0,
        model: BaseModel = Webhook,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "POST"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/{base_name}/webhooks/"
        JSON = {"url": url, "secret": str(secret)}
        ITEM = "webhook"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            results = response[ITEM]

        if model:
            results = model(**results)

        return results

    # [WEBHOOKS] update webhook
    async def update_webhook(
        self,
        workspace_name_or_id: Union[str, int],
        base_name: str,
        webhook_id: str,
        url: str,
        secret: int = None,
    ):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "PUT"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/{base_name}/webhooks/{webhook_id}/"
        JSON = {"url": url, "secret": str(secret)}

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL, json=JSON)
            print(response)

    # [WEBHOOKS] delete webhook
    async def delete_webhook(self, workspace_name_or_id: Union[str, int], base_name: str, webhook_id: str):
        workspace = await self.get_workspace(name_or_id=workspace_name_or_id)

        METHOD = "DELETE"
        URL = f"/api/v2.1/workspace/{workspace.id}/dtable/{base_name}/webhooks/{webhook_id}/"
        ITEM = "success"

        async with self.session_maker(token=self.account_token) as session:
            response = await self.request(session=session, method=METHOD, url=URL)
            results = response[ITEM]

        return results

    ################################################################
    # SNAPSHOTS
    ################################################################
    # TBD



import logging
from datetime import date, datetime
from typing import List

from ...const import DT_FMT, TZ
from ...model import Column, Table, User, Metadata
from .deserializer import ColumnDeserializer, Deserializer

logger = logging.getLogger(__name__)


################################################################
# Python Types for SeaTable
################################################################
class PythonCheckbox(ColumnDeserializer):
    def schema(self):
        return bool

    def convert(self, x):
        return bool(x)


class PythonText(ColumnDeserializer):
    def schema(self):
        return str

    def convert(self, x):
        return str(x)


class PythonRate(ColumnDeserializer):
    def schema(self):
        return int

    def convert(self, x):
        return int(x)


class _PythonInteger(ColumnDeserializer):
    def schema(self):
        return int

    def convert(self, x):
        return int(x)


class _PythonFloat(ColumnDeserializer):
    def schema(self):
        return float

    def convert(self, x):
        return float(x)


class PythonNumber(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )
        if self.data.get("enable_precision") and (self.data.get("precision") == 0):
            self.sub_deserializer = _PythonInteger(name=self.name, seatable_type=self.seatable_type, data=self.data)
        else:
            self.sub_deserializer = _PythonFloat(name=self.name, seatable_type=self.seatable_type, data=self.data)

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        return self.sub_deserializer(x)


class _PythonDate(ColumnDeserializer):
    def schema(self):
        return date

    def convert(self, x):
        return date.fromisoformat(x[:10])


class _PythonDatetime(ColumnDeserializer):
    def schema(self):
        return datetime

    def convert(self, x):
        if x.endswith("Z"):
            x = x.replace("Z", "+00:00", 1)
        try:
            x = datetime.strptime(x, DT_FMT)
        except Exception as ex:
            x = datetime.fromisoformat(x)
        return x.astimezone(TZ)


class PythonDate(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )
        if self.data and self.data["format"] == "YYYY-MM-DD":
            self.sub_deserializer = _PythonDate(name=self.name, seatable_type=self.seatable_type, data=self.data)
        else:
            self.sub_deserializer = _PythonDatetime(name=self.name, seatable_type=self.seatable_type, data=self.data)

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        return self.sub_deserializer(x)


class PythonDuration(ColumnDeserializer):
    def schema(self):
        return int

    def convert(self, x):
        return x


class PythonSingleSelect(ColumnDeserializer):
    def schema(self):
        return str

    def convert(self, x):
        return str(x)


class PythonMulitpleSelect(ColumnDeserializer):
    def schema(self):
        return List[str]

    def convert(self, x):
        return [str(_x) for _x in x]


class PythonUser(ColumnDeserializer):
    def schema(self):
        return str

    def convert(self, x):
        if not self.collaborator_map:
            return x
        return self.collaborator_map[x] if x in self.collaborator_map else x


class PythonListUsers(ColumnDeserializer):
    def schema(self):
        return List[str]

    def convert(self, x):
        if not self.collaborator_map:
            return x
        return [self.collaborator_map[_x] if _x in self.collaborator_map else _x for _x in x]


class PythonFile(ColumnDeserializer):
    def schema(self):
        return List[str]

    def convert(self, x):
        return [_x["url"] for _x in x]


class PythonImage(ColumnDeserializer):
    def schema(self):
        return List[str]

    def convert(self, x):
        return x


class PythonAutoNumber(ColumnDeserializer):
    def schema(self):
        return str

    def convert(self, x):
        return str(x)


DESERIALIZER = {
    "row-id": PythonText,
    "checkbox": PythonCheckbox,
    "text": PythonText,
    "long-text": PythonText,
    "string": PythonText,  # [NOTE] formula column의 result_type이 'text'가 아닌 'string'을 반환.
    "button": PythonText,
    "email": PythonText,
    "url": PythonText,
    "rate": PythonRate,
    "number": PythonNumber,
    "date": PythonDate,
    "duration": PythonDuration,
    "ctime": PythonDate,
    "mtime": PythonDate,
    "single-select": PythonSingleSelect,
    "multiple-select": PythonMulitpleSelect,
    "user": PythonUser,
    "collaborator": PythonListUsers,
    "creator": PythonUser,
    "last-modifier": PythonUser,
    "file": PythonFile,
    "image": PythonImage,
    "auto-number": PythonAutoNumber,
}


class PythonFormula(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )
        self.sub_deserializer = DESERIALIZER[self.data["result_type"]](
            name=self.name, seatable_type=self.data["result_type"], data=dict()
        )

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        if x == "#VALUE!":
            return None
        return self.sub_deserializer(x)


class PythonLink(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )

        if "array_type" in data:
            array_type = self.data["array_type"]
            array_data = self.data["array_data"]
        else:
            column = self.get_column_by_id(table_id=data["table_id"], column_id=data["display_column_key"])
            array_type = column.type
            array_data = column.data

        self.sub_deserializer = DESERIALIZER[array_type](name=self.name, seatable_type=array_type, data=array_data)
        self.is_multiple = self.data["is_multiple"]

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        x = [self.sub_deserializer(_x.get("display_value")) for _x in x]
        if not x:
            return None
        if self.is_multiple:
            return x
        return x[0]


class PythonLinkFormula(ColumnDeserializer):
    def __init__(
        self,
        name: str,
        seatable_type: str,
        data: dict = None,
        metadata: Metadata = None,
        collaborator_map: dict = None,
    ):
        super().__init__(
            name=name, seatable_type=seatable_type, data=data, metadata=metadata, collaborator_map=collaborator_map
        )

        result_type = self.data["result_type"]

        # [TODO] is_multiple 추가해서 참조하는 link가 리스트이면 리스트, 단일 값이면 단일 값을 반환하도록
        if result_type == "array":
            array_type = self.data["array_type"]
            array_data = self.data["array_data"]
            self.sub_deserializer = DESERIALIZER[array_type](name=self.name, seatable_type=array_type, data=array_data)

        if result_type == "number":
            result_data = {}
            if self.data["formula"] == "count_links":
                result_data.update({"enable_precision": True, "precision": 0})
            self.sub_deserializer = DESERIALIZER[result_type](
                name=self.name, seatable_type=result_type, data=result_data
            )

    def schema(self):
        return self.sub_deserializer.schema()

    def convert(self, x):
        if isinstance(x, list):
            return [self.sub_deserializer(_x) for _x in x]
        return self.sub_deserializer(x)


DESERIALIZER.update(
    {
        "formula": PythonFormula,
        "link": PythonLink,
        "link-formula": PythonLinkFormula,
    }
)


################################################################
# Seatable To Python Deserializer
################################################################
class ToPython(Deserializer):
    Deserializer = DESERIALIZER

    def schema(self):
        return {
            "name": self.generate_table_name(),
            "columns": [{name: column.schema()} for name, column in self.columns.items()],
        }



from datetime import datetime
import sqlparse

from .const import DT_FMT, TZ


# divide chunks
def divide_chunks(x: list, chunk_size: int):
    for i in range(0, len(x), chunk_size):
        yield x[i : i + chunk_size]


# parse string datetime
def parse_str_datetime(x):
    if x.endswith("Z"):
        x = x.replace("Z", "+00:00", 1)
    try:
        x = datetime.strptime(x, DT_FMT)
    except Exception:
        x = datetime.fromisoformat(x)
    return x.astimezone(TZ)


# extract table name error
class NoTableName(Exception):
    def __init__(self, msg):
        super().__init__(msg)


# extract table name
def extract_table_name(sql):
    statements = sqlparse.parse(sql)
    statement = statements[0]

    # FROM 절을 찾습니다.
    from_seen = False
    for token in statement.tokens:
        if from_seen:
            if isinstance(token, sqlparse.sql.IdentifierList):
                for identifier in token.get_identifiers():
                    return identifier.get_real_name()
            if isinstance(token, sqlparse.sql.Identifier):
                return token.get_real_name()
        if token.ttype is sqlparse.tokens.Keyword and token.value.upper() == "FROM":
            from_seen = True
    else:
        _msg = f"no table name found '{sql}'!"
        raise NoTableName(_msg)


# extract columns
def extract_columns_from_select(sql):
    statements = sqlparse.parse(sql)
    statement = statements[0]

    select_seen = False
    columns = []
    for token in statement.tokens:
        if token.ttype is sqlparse.tokens.DML and token.value.upper() == "SELECT":
            select_seen = True
        if select_seen:
            if isinstance(token, sqlparse.sql.IdentifierList):
                for identifier in token.get_identifiers():
                    columns.append(identifier.get_real_name())
            elif isinstance(token, sqlparse.sql.Identifier):
                columns.append(token.get_real_name())
            elif token.ttype is sqlparse.tokens.Keyword and token.value.upper() == "FROM":
                break
    return columns



from datetime import datetime
from typing import List

from .core import _Model

__all__ = ["Activity"]


class Activity(_Model):
    dtable_uuid: str  # "d48c3aae-dbc1-4325-a890-1c2e79ea5319"
    workspace_id: int  # 125
    dtable_name: str  # "SeaTable API Docs"
    dtable_icon: str = None  # "icon-research"
    dtable_color: str = None  # "#1688FC"
    op_date: datetime  # "2020-11-09T09:59:13+00:00"
    insert_row: int = None  # 0
    modify_row: int = None  # 13
    delete_row: int = None  # 0


class ColumnOption(_Model):
    name: str
    color: str
    textColor: str
    id: str


class Column(_Model):
    options: List[ColumnOption]
    value: str = None
    old_value: str = None


class Row(_Model):
    column_key: str
    column_name: str
    column_type: str
    column_data: Column


class Data(_Model):
    dtable_uuid: str
    row_id: str
    op_user: str
    op_type: str
    op_time: float
    table_id: str
    table_name: str
    row_name: str = None
    row_data: List[Row] = None
    op_app: str = None


class Event(_Model):
    event: str
    data: Data



import parse
import pyarrow as pa

from ...const import SYSTEM_FIELDS
from ...model.column import (
    AutoNumber,
    Button,
    Checkbox,
    Collaborator,
    CreationTime,
    Creator,
    Date,
    Datetime,
    Duration,
    Email,
    File,
    Formula,
    Image,
    Integer,
    LastModificationTime,
    LastModifier,
    Link,
    LinkFomula,
    LongText,
    MultipleSelect,
    Number,
    Rate,
    SingleSelect,
    Text,
    Url,
)

# Arrow to Seatable Schema
SCHEMA_MAP = {
    "bool": Checkbox,
    "int8": Integer,
    "int16": Integer,
    "int32": Integer,
    "int64": Integer,
    "uint8": Integer,
    "uint16": Integer,
    "uint32": Integer,
    "uint64": Integer,
    "halffloat": Number,  # float16
    "float": Number,  # float32
    "double": Number,  # float64
    "time32": Duration,
    "time64": Duration,
    "timestamp": Datetime,
    "date32": Date,
    "date64": Date,
    "duration": Duration,
    "string": Text,
    "utf8": Text,
    "large_string": LongText,
    "large_utf8": LongText,
    "decimal128": Number,
    "list": MultipleSelect,
    "large_list": MultipleSelect,
}


ARROW_STR_DTYPE_PATTERNS = [
    parse.compile("{dtype}[{unit}, tz={tz}]"),
    parse.compile("{dtype}[{unit}]"),
    parse.compile("{dtype}({precision}, {scale})"),
    parse.compile("{dtype}<item: {item}>"),
    parse.compile("{dtype}<item: {item}>[{list_size}]"),
    parse.compile("{dtype}"),
]


class FromArrowTable:
    def __init__(self, tbl: pa.Table):
        self.tbl = tbl
        self._schema = [(c, str(tbl.schema.field(c).type)) for c in tbl.schema.names]

        # get deserializer opts
        self.opts = {column: self.dtype_parser(dtype) for column, dtype in self._schema}

        # seatable schema
        self.columns = [
            SCHEMA_MAP[opt["dtype"]](name=name) for name, opt in self.opts.items() if name not in SYSTEM_FIELDS
        ]

    @staticmethod
    def dtype_parser(x):
        for pattern in ARROW_STR_DTYPE_PATTERNS:
            r = pattern.parse(x)
            if r:
                return r.named

    def get_rows_for_append(self):
        return [{k: v for k, v in r.items() if k not in SYSTEM_FIELDS} for r in self.tbl.to_pylist()]

    def get_rows_for_update(self, row_id_field: str = "_id"):
        updates = list()
        for row in self.tbl.to_pylist():
            updates.append(
                {
                    "row_id": row[row_id_field],
                    "row": {k: v for k, v in row.items() if k not in SYSTEM_FIELDS},
                }
            )
        return updates

    def null(self, value):
        pass

    def bool(self, value):
        pass

    def int8(self, value):
        pass

    def int16(self, value):
        pass

    def int32(self, value):
        pass

    def int64(self, value):
        pass

    def uint8(self, value):
        pass

    def uint16(self, value):
        pass

    def uint32(self, value):
        pass

    def uint64(self, value):
        pass

    def halffloat(self, value):
        pass

    def float(self, value):
        pass

    def double(self, value):
        pass

    def time32(self, value, unit):
        pass

    def time64(self, value, unit):
        pass

    def timestamp(self, value, unit, tz: str = None):
        pass

    def date32(self, value, unit: str):
        pass

    def date64(self, value, unit: str):
        pass

    def duration(self, value, unit: str):
        pass

    def string(self, value):
        pass

    def large_string(self, value):
        pass

    def decimal128(self, value, precision: int = 0, scale: int = 0):
        pass

    def list(self, value, item, list_size: int = -1):
        pass

    def large_list(self, value, item, list_size: int = -1):
        pass



import os

from dotenv import load_dotenv

load_dotenv()

# Seatable
SEATABLE_URL = os.getenv("SEATABLE_URL")
SEATABLE_USERNAME = os.getenv("SEATABLE_USERNAME")
SEATABLE_PASSWORD = os.getenv("SEATABLE_PASSWORD")
SEATABLE_ACCOUNT_TOKEN = os.getenv("SEATABLE_ACCOUNT_TOKEN")
SEATABLE_API_TOKEN = os.getenv("SEATABLE_API_TOKEN")
